from karen.evaluate import evaluate
from karen.classify import CLASSIFICATIONS

COMBO_SEQUENCES = {}

OVERRIDE_COMBO_SEQUENCES = {
    "Fast Panther" : "tGslptu",
    "Reverse Yo-Yo" : "tgdotu",
    "Master Manipulator" : "otslptu",
    "Further Beyond" : "otuwtotgwtuwtodto",
    "Yo-Yo" : "gdwtuwto",
    "Botched Yo-Yo" : "gdwtuototu",
    "Agni-Kai Yo-Yo" : "gdtutp",
    "Bald Slam" : "twGuot",
    "Vortex" : "uwtdGo",
    "In And Out" : "tbuwtg",
    "Hydro Combo" : "tgptalptu"
}

COMBO_ALIASES = {
    "bnb" : "Bread & Butter (BnB)",
    "bnbplink" : "BnB Long Plink",
    "fishing" : "Fishing Combo / Sekkombo",
    "fish" : "Fishing Combo / Sekkombo",
    "sekkombo" : "Fishing Combo / Sekkombo",
    "sekombo" : "Fishing Combo / Sekkombo",
    "gripkickrip" : "Grip Kick Rip (GKR)",
    "gkr" : "Grip Kick Rip (GKR)",
    "ohburst" : "Overhead Burst",
    "fantastic" : "Fantastic Killer",
    "sapstack" : "Saporen FFAmestack",
    "agnikai" : "Agni-Kai Yo-Yo",
    "bald" : "Bald Slam",
    "skypull" : "Yo-Yo",
    "spc" : "Yo-Yo",
    "skyyoink" : "Yo-Yo",

    "burnbnb" : "Burn BnB / Fadeaway",
    "fadeaway" : "Burn BnB / Fadeaway",
    "burnohburst" : "Burn Overhead Burst",
    "burnoverhead" : "Burn Overhead Burst",
    "burnoh" : "Burn Overhead Burst",
    "friedfish" : "Fried Fish / Firehook",
    "fried" : "Fried Fish / Firehook",
    "burnsekkombo" : "Fried Fish / Firehook",
    "burnsekombo" : "Fried Fish / Firehook",
    "firehook" : "Fried Fish / Firehook",
    "innout" : "In And Out",
    "in&out" : "In And Out"
}

def loadComboSequences():
    for sequence in CLASSIFICATIONS:
        COMBO_SEQUENCES[CLASSIFICATIONS[sequence]] = sequence
    for combo in OVERRIDE_COMBO_SEQUENCES:
        COMBO_SEQUENCES[combo] = OVERRIDE_COMBO_SEQUENCES[combo]

    for sequence in CLASSIFICATIONS:
        filterName = CLASSIFICATIONS[sequence].replace(" ", "").replace("-", "").lower()
        if len(filterName) > 5 and filterName[-5:] == "combo":
            filterName = filterName[:-5]
        COMBO_ALIASES[filterName] = CLASSIFICATIONS[sequence]

    for name in COMBO_ALIASES.copy():
        if "bnb" in name:
            COMBO_ALIASES[name.replace("bnb", "b&b")] = COMBO_ALIASES[name]
            COMBO_ALIASES[name.replace("bnb", "bandb")] = COMBO_ALIASES[name]
            COMBO_ALIASES[name.replace("bnb", "breadnbutter")] = COMBO_ALIASES[name]
            COMBO_ALIASES[name.replace("bnb", "bread&butter")] = COMBO_ALIASES[name]
            COMBO_ALIASES[name.replace("bnb", "breadandbutter")] = COMBO_ALIASES[name]


def getCombo(name):
    if len(COMBO_SEQUENCES) == 0:
        loadComboSequences()

    filterName = name.replace(" ", "").replace("-", "").lower()
    if len(filterName) > 5 and filterName[-5:] == "combo":
            filterName = filterName[:-5]

    if not filterName in COMBO_ALIASES or not COMBO_ALIASES[filterName] in COMBO_SEQUENCES:
        return "```\nERROR: Combo not found\n```"
    
    return evaluate(COMBO_SEQUENCES[COMBO_ALIASES[filterName]])
        
def listCombos():
    comboList = []
    sequenceList = []
    maxLength = 0

    for sequence in CLASSIFICATIONS:
        if not CLASSIFICATIONS[sequence] in comboList:
            comboList += [CLASSIFICATIONS[sequence]]
            sequenceList += [sequence]
            maxLength = max(maxLength, len(comboList[-1]))

    return "```\n" + "\n".join([comboList[i] + " " * (maxLength - len(comboList[i])) + " | " + sequenceList[i] for i in range(len(comboList))]) + "\n```"


from karen.actions import *

class Charge:
    MAX_CHARGES = 0
    currentCharge = 0
    RECHARGE_TIME = 0
    COOLDOWN_TIME = 0
    cooldownTimer = 0
    activeTimer = 0 # cooldown refresh= doesn't start until icon is no longer yellow

    def __init__(self, MAX_CHARGES, RECHARGE_TIME, COOLDOWN_TIME = 0):
        self.MAX_CHARGES = MAX_CHARGES
        self.currentCharge = MAX_CHARGES * RECHARGE_TIME
        self.RECHARGE_TIME = RECHARGE_TIME
        self.COOLDOWN_TIME = COOLDOWN_TIME


class State:

    # charges/cooldowns - to avoid floating point innacuracy, using an ability takes charge equal to the number of frames it will take to recharge
    charges = {}
    tracerActiveTimer = 0 # frames remaining until tracer on opponent expires
    burnTracerActiveTimer = 0
    burnActiveTimer = 0 # timer for while burn tracer deals damage after procced
    gohtWaitTime = 0 # countdown from when tracer fired at unmarked target to when goht can be used

    maxPossibleRange = 24 # for lowering the upper bound when factoring in travel time

    # airborne
    isAirborn = False

    # overheads
    hasDoubleJump = True
    hasJumpOverhead = False
    hasSwingOverhead = False
    removeSwingOnEnd = False
    swingWaitTimer = 0 # frames until swing/whiff can be used after autoswing

    # punch sequence
    punchSequence = 0 # 0 & 1 correspond to punches, 2 corresponds to kick
    punchSequenceTimer = 0 # frames remaining until punch sequence resets
    punchWaitTimer = 0 # frames until a punch can be used

    # seasonal boost
    damageMultiplier = 1

    # tracking metrics
    damageDealt = 0
    timeTaken = 0
    firstDamageTime = 0
    timeFromDamage = 0 # only computed when resolve() is called
    minTimeTaken = 0 # makes sure swing cancels at the end don't cut the timer short

    # sequence output
    sequence = ""

    def __init__(self):

        self.charges = {
            "t" : Charge(
                MAX_CHARGES = 5, 
                RECHARGE_TIME = 2.5 * 60,
                COOLDOWN_TIME = 0.5 * 60
            ),
            "s" : Charge(
                MAX_CHARGES = 3, 
                RECHARGE_TIME = 6 * 60
            ),
            "g" : Charge(
                MAX_CHARGES = 1, 
                RECHARGE_TIME = 8 * 60
            ),
            "u" : Charge(
                MAX_CHARGES = 2, 
                RECHARGE_TIME = 6 * 60,
                COOLDOWN_TIME = 2 * 60
            ),
            "b" : Charge(
                MAX_CHARGES = 1, 
                RECHARGE_TIME = 12 * 60
            )
        }

    def incrementTime(self, frames, warnings):

        if frames <= 0:
            return
        
        self.timeTaken += frames

        for chargeType in self.charges:
            self.charges[chargeType].currentCharge = min(self.charges[chargeType].currentCharge + frames, self.charges[chargeType].MAX_CHARGES * self.charges[chargeType].RECHARGE_TIME)
            self.charges[chargeType].cooldownTimer = max(self.charges[chargeType].cooldownTimer - frames, 0)

            if self.charges[chargeType].activeTimer > 0 and self.charges[chargeType].activeTimer <= frames:
                excessTime = frames - self.charges[chargeType].activeTimer
                self.endAction(chargeType)
                self.charges[chargeType].currentCharge += excessTime
            
            elif self.charges[chargeType].activeTimer > 0:
                self.charges[chargeType].activeTimer -= frames

        if self.tracerActiveTimer > 0 and self.tracerActiveTimer <= frames:
           warnings += ["tracer expired without proc after " + self.sequence]
        if self.burnTracerActiveTimer > 0 and self.burnTracerActiveTimer <= frames:
           warnings += ["burn tracer expired without proc after " + self.sequence]    

        # burn tracer damage ticks 5 times per second (12 frames)
        burnFrames = frames
        if self.burnActiveTimer > BURN_TRACER_BURN_TIME:
            if burnFrames + BURN_TRACER_BURN_TIME >= self.burnTracerActiveTimer:
                burnFrames -= self.burnActiveTimer - BURN_TRACER_BURN_TIME
                self.burnActiveTimer = BURN_TRACER_BURN_TIME
                self.damageDealt += BURN_TRACER_DPS / 5
            else:
                self.burnActiveTimer -= burnFrames
                burnFrames = 0
        while burnFrames >= 12 and self.burnActiveTimer >= 12:
            burnFrames -= 12
            self.burnActiveTimer -= 12
            if self.burnActiveTimer != 0:
                self.damageDealt += BURN_TRACER_DPS / 5
        if burnFrames > self.burnActiveTimer:
            self.burnActiveTimer = 0
        elif burnFrames > 0:
            if (self.burnActiveTimer % 12) <= burnFrames and (self.burnActiveTimer % 12) != 0  and self.burnActiveTimer > 12:
                self.damageDealt += BURN_TRACER_DPS / 5
            self.burnActiveTimer -= burnFrames

        self.tracerActiveTimer = max(self.tracerActiveTimer - frames, 0)
        self.burnTracerActiveTimer = max(self.burnTracerActiveTimer - frames, 0)
        self.gohtWaitTime = max(self.gohtWaitTime - frames, 0)

        self.punchWaitTimer = max(0, self.punchWaitTimer - frames)
        self.swingWaitTimer = max(0, self.swingWaitTimer - frames)


        self.punchSequenceTimer = max(self.punchSequenceTimer - frames, 0)
        if self.punchSequenceTimer == 0:
            self.punchSequence = 0

    # if not all damage has actually taken place, await the final damage ticks
    def resolve(self, warnings=[]):
        self.incrementTime(self.minTimeTaken - self.timeTaken, warnings)
        self.timeFromDamage = self.timeTaken - self.firstDamageTime

    def endAction(self, action): # ends current action and takes away the associated cooldown charge
        self.charges[action].activeTimer = 0
        if action != "s" or self.removeSwingOnEnd:
            self.charges[action].currentCharge -= self.charges[action].RECHARGE_TIME
        if action == "s":
            self.hasSwingOverhead |= self.isAirborn and self.hasDoubleJump
            self.hasJumpOverhead |= self.isAirborn and not self.hasDoubleJump

    def inferInitialState(self, comboSequence, warnings=[]):
        foldSequence = "".join(comboSequence)

        # pre-tag if GOHT used before tracer / burn tracer
        if "G" in foldSequence and ((not ("t" in foldSequence or "b" in foldSequence)) or foldSequence.index("G") < min((foldSequence+"bt").index("t"), (foldSequence+"bt").index("b"))):
            self.tracerActiveTimer = TRACER_ACTIVE_TIME
            warnings += ["inferred target starts with tracer applied to enable GOHT"]
        
        # pre-punch if kick is used before two punches
        if "k" in foldSequence:
            priorPunches = foldSequence.count("p", 0, foldSequence.index("k"))
            self.punchSequence = max(0, 2 - priorPunches)
            self.punchSequenceTimer = PUNCH_SEQUENCE_MAX_DELAY - ACTIONS["p"].cancelTimes[foldSequence.replace("j", "").replace("l", "")[0]] # set punch sequence timer as if pre-punching was the last thing done before the combo
        
        # airborne if overhead/land is used before jump/swing/uppercut/burn tracer
        temp = foldSequence + "jdsub" # ensures index funtions don't error
        preAirborne = temp[:min(temp.index("j"), temp.index("s"), temp.index("d"), temp.index("u"), temp.index("b"))]
        self.isAirborn = "o" in preAirborne or "l" in preAirborne

        # also airborne if only one jump is used before overhead
        if "o" in foldSequence and not self.isAirborn:
            preOH = foldSequence[:temp.index("o")]
            self.isAirborn = not ("s" in preOH or "w" in preOH or "b" in preOH or preOH.count("j") >= 2)

        # has swing overhead if overhead is used before payout
        temp = foldSequence + "jjdswb"
        if not self.isAirborn: # ignore the first jump for overhead payouts if not airborne
            temp = temp[:temp.index("j")] + temp[temp.index("j") + 1:]
        preOverheadPayout = temp[:min(temp.index("j"), temp.index("d"), temp.index("s"), temp.index("w"), temp.index("b"))]
        self.hasSwingOverhead = preOverheadPayout.count("o") >= 1

        # has jump ovehead if two overheads are used before payout
        if preOverheadPayout.count("o") >= 2:
            self.hasJumpOverhead = True
            self.hasDoubleJump = False

        # infers jump overhead being preserved through goht
        if "G" in preOverheadPayout and (not ("sG" in preOverheadPayout or "wG" in preOverheadPayout)) and "o" in preOverheadPayout[preOverheadPayout.index("G"):]:
            self.hasJumpOverhead = True
            self.hasDoubleJump = False

        # infers starting within 5m if melee is used before goht/goh
        preGOH = foldSequence + "gG"
        preGOH = preGOH[:min(preGOH.index("G"), preGOH.index("g"))]
        for a in preGOH:
            if a in "pkou":
                self.maxPossibleRange = 5
                break
        

    def correctSequence(self, combo):

        # Converts relevant jumps to double jumps
        isAirborne = self.isAirborn
        correctedCombo = []
        for action in combo:
            if isAirborne and action == "j":
                correctedCombo += ["d"]
            else:
                correctedCombo += [action]
            
            if action in ["j", "s", "d", "u", "b"]:
                isAirborne = True
            elif action == "l":
                isAirborne = False
        
        # Corrects action sequences
        temp = correctedCombo
        correctedCombo = []
        i = 0
        while i in range(len(temp)):

            # space jam
            if i+2 < len(temp) and ("".join(temp[i:i+3]) in ["uwG", "usG"]):
                correctedCombo += ["u+w+G"]
                i += 3
                continue
                
            if i+1 < len(temp):
                sequence = "".join(temp[i:i+2])
                
                # space jam
                if sequence == "uG":
                    correctedCombo += ["u+w+G"]
                    i += 2
                    continue

                # double jump
                if sequence == "jd":
                    correctedCombo += ["d"]
                    i += 2
                    continue
                
                if i > 0 and temp[i] == "d" and temp[i - 1] in "jd":
                    i += 1
                    continue

                # saporens
                if sequence in ["pG", "kG", "oG"]:
                    correctedCombo += [sequence[0] + "+G"]
                    i += 2
                    continue

            correctedCombo += [temp[i]]
            i += 1
        
        return correctedCombo


import copy

PUNCH_SEQUENCE_MAX_DELAY = 90 # punch/kick must start <90 frames after last punch to advance the sequence
TRACER_ACTIVE_TIME = 4 * 60
TRACER_PROC_DAMAGE = 45
TRACER_MAX_TRAVEL_TIME = 10

BURN_TRACER_ACTIVE_TIME = 3 * 60
BURN_TRACER_DPS = 15
BURN_TRACER_BURN_TIME = 240
BURN_TRACER_MAX_TRAVEL_TIME = 3
BURN_TRACER_BACKFLIP_DISTANCE = 7


ACTION_NAMES = {
    "j" : "j", "jump" : "j",
    "d" : "d", "double jump" : "d", "dj" : "d",
    "l" : "l", "land" : "l",
    "p" : "p", "punch" : "p", "puncha" : "p",  "punchb" : "p", "meleepunch" : "p",  "meleepuncha" : "p",  "meleepunchb" : "p",
    "k" : "k", "kick" : "k",  "meleekick" : "k",
    "o" : "o", "overheadslam" : "o",  "overhead" : "o", "over" : "o", "oh" : "o",  "meleeoverhead" : "o",  "slam" : "o",
    "t" : "t", "tracer" : "t",  "webtracer" : "t",  "cluster" : "t",  "webcluster" : "t",
    "s" : "s", "swing" : "s", "webswing" : "s",  "highswing" : "s",  "lowswing" : "s", "z" : "s", "zip" : "s", "webzip" : "s",
    "a" : "a", "automatic swing" : "a", "autoswing" : "a", "automatic" : "a", "auto" : "a", "easyswing" : "a", "easy" : "a", "simple swing" : "a", "simple" : "a",
    "w" : "w", "whiff" : "w", "webwhiff" : "w", "swingwhiff" : "w", "wiff" : "w", "web wiff" : "w", "swing wiff" : "w",
    "g" : "g", "getoverhere" : "g", "goh" : "g", "webpull" : "g", "pull" : "g",
    "G" : "G", "getoverheretargeting" : "G", "getoverheretargetting" : "G", "goht" : "G",
    "u" : "u", "uppercut" : "u", "upper" : "u", "amazingcombo" : "u",
    "b" : "b", "burn" : "b", "burntracer" : "b", "burncluster" : "b", "fire" : "b", "firetracer" : "b", "firecluster" : "b", "flame" : "b", "flametracer" : "b", "flamecluster" : "b",
    "+" : "+", "/" : "+",

    "G+u" : "G+u", "f" : "G+u", "ffame" : "G+u", "ffamestack" : "G+u",
    "o+G" : "o+G", "saporen" : "o+G", "sap" : "o+G", "overheadsaporen" : "o+G", "ohsap" : "o+G",
    "p+G" : "p+G", "punchsaporen" : "p+G", "punchsap" : "p+G",
    "k+G" : "k+G","kicksaporen" : "k+G", "kicksap" : "k+G",
    "o+G+u" : "o+G+u", "F" : "o+G+u", "saporenffamestack" : "o+G+u", "sapffamestack" : "o+G+u", "sapffame" : "o+G+u", "overheadsaporenffamestack" : "o+G+u", "ohsapffamestack" : "o+G+u", "ohsapffame" : "o+G+u",
    "p+G+u" : "p+G+u", "punchsaporenffamestack" : "p+G+u", "punchsapffamestack" : "p+G+u", "punchsapffame" : "p+G+u",
    "k+G+u" : "k+G+u", "kicksaporenffamestack" : "k+G+u", "kicksapffamestack" : "k+G+u", "kicksapffame" : "k+G+u",
    "u+w+G" : "u+w+G", "J" : "u+w+G", "spacejam" : "u+w+G", "sj" : "u+w+G", "u+s+G" : "u+w+G", "u+G" : "u+w+G",
    "p+t" : "p+t", "r" : "p+t", "reversetrigger" : "p+t", "rt" : "p+t", "backflash" : "p+t", "punchreversetrigger" : "p+t", "punchbackflash" : "p+t", "punchrt" : "p+t", "punchblackflash" : "p+t", "blackflash" : "p+t",
    "k+t" : "k+t", "kickreversetrigger" : "k+t", "kickbackflash" : "k+t", "kickrt" : "k+t", "jashflash" : "k+t", "kickblackflash" : "k+t",
    "o+t" : "o+t", "overheadreversetrigger" : "o+t", "overheadbackflash" : "o+t", "overheadrt" : "o+t", "ohreversetrigger" : "o+t", "ohbackflash" : "o+t", "ohrt" : "o+t", "overheadblackflash" : "o+t", "ohblackflash" : "o+t",
    "p+o" : "p+o", "punchoverheadstack" : "p+o", "punchohstack" : "p+o", "unique3hitpunchstack" : "p+o", "uniquethreehitpunchstack" : "p+o","unique3hitpunch" : "p+o", "uniquethreehitpunch" : "p+o", "u3hpunchstack" : "p+o", "u3hpunch" : "p+o",
    "k+o" : "k+o", "kickoverheadstack" : "k+o", "kickohstack" : "k+o", "unique3hitkickstack" : "k+o", "uniquethreehitkickstack" : "k+o","unique3hitkick" : "k+o", "uniquethreehitkick" : "k+o", "u3hkickstack" : "k+o", "u3hkick" : "k+o",

    "shortplink" : "so",
    "longplink" : "wto", "plink" : "wto",
    "breadandbutter" : "tGu", "breadnbutter" : "tGu", "bnb" : "tGu",
    "uniquethreehit" : "wpp+o", "unique3hit" : "wpp+o", "u3h" : "wpp+o",
    "slytech" : "suo", "sly" : "suo"
}


class Action:
    name = ""

    damage = 0
    procsTracer = False
    procTime = 0

    damageTime = 0 # number of frames between activation and all damage registering
    firstDamageTime = 0 # for movestacks with multiple hits
    maxTravelTime = 0 # max number of added frames that projectile actions can take to travel
    range = 0 # purely for travel time calculations
    cancelTimes = {} # dict maps action to how many frames into this actions' animation the given action can be used
    
    chargeActivations = {} # the cooldown charges used and how long before the recharge starts from the action start time (for activeTimer)
    endActivations = {} # some abilities cause others to start recharging immediately
    awaitCharges = {} # which cooldown charges have to be ready to use during the action, and how long into the action they're used (useful for stacks)

    def __init__(self, name, damage = 0, procsTracer=False, procTime = 0, damageTime = 0, firstDamageTime = 0, maxTravelTime = 0, range = 0, cancelTimes={}, chargeActivations={}, endActivations={}, awaitCharges={}):
        self.name = name
        self.damage = damage
        self.procsTracer = procsTracer
        self.procTime = damageTime if procTime == 0 else procTime
        self.damageTime = damageTime
        self.firstDamageTime = firstDamageTime if firstDamageTime != 0 else damageTime
        self.maxTravelTime = maxTravelTime
        self.range = range
        self.cancelTimes = cancelTimes
        self.chargeActivations = chargeActivations
        self.cancelTimes[""] = damageTime # time to transition to the end of the combo is damageTime
        self.endActivations = endActivations
        self.awaitCharges = awaitCharges


ACTIONS = {

    "p" : Action (
        name = "Punch",
        damage = 25, 
        procsTracer = True, 
        
        damageTime = 14,
        range = 3,
        cancelTimes = {
            "p" : 23,
            "k" : 23,
            "o" : 23,
            "t" : 11,
            "s" : 11,
            "a" : 11,
            "w" : 11,
            "g" : 11,
            "G" : 11,
            "u" : 11,
            "b" : 11
        }
    ),

    "k" : Action (
        name = "Kick",
        damage = 40, 
        procsTracer = True,
        
        damageTime = 24,
        range = 4,
        cancelTimes = {
            "p" : 49,
            "k" : 49,
            "o" : 49,
            "t" : 22,
            "s" : 22,
            "a" : 22,
            "w" : 22,
            "g" : 22,
            "G" : 22,
            "u" : 22,
            "b" : 22
        }
    ),

    "o" : Action (
        name = "Overhead",
        damage = 55, 
        procsTracer = True, 
        
        damageTime = 38,
        range = 4,
        cancelTimes = {
            "p" : 53,
            "k" : 53,
            "o" : 53,
            "t" : 33,
            "s" : 33,
            "a" : 33,
            "w" : 33,
            "g" : 33,
            "G" : 33,
            "u" : 33,
            "b" : 33
        }
    ),

    "t" : Action (
        name = "Tracer",
        damage = 30, 
        
        damageTime = 12,
        range = 20,
        cancelTimes = {
            "p" : 30,
            "k" : 30,
            "o" : 30,
            "t" : 30,
            "s" : 6,
            "a" : 6,
            "w" : 6,
            "g" : 6,
            "G" : 6,
            "u" : 6,
            "b" : 6
        },

        chargeActivations = { "t" : 0 },
        endActivations = ["s"],
        awaitCharges = { "t" : 0 }
    ),

    "s" : Action (
        name = "Swing",
        
        maxTravelTime = 7,
        range = 30,
        cancelTimes = {
            "p" : 10,
            "k" : 10,
            "o" : 10,
            "t" : 10,
            "s" : 10,
            "a" : 10,
            "w" : 10,
            "g" : 1,
            "G" : 1,
            "u" : 1,
            "b" : 1
        },

        chargeActivations = { "s" : 10 },
        endActivations = ["g", "u", "b"],
        awaitCharges = { "s" : 0 }
    ),

    "a" : Action (
        name = "Auto Swing",
        
        cancelTimes = {
            "p" : 1,
            "k" : 1,
            "o" : 1,
            "t" : 1,
            "s" : 19,
            "a" : 19,
            "w" : 19,
            "g" : 1,
            "G" : 1,
            "u" : 1,
            "b" : 1
        },

        chargeActivations = { "s" : 1 },
        endActivations = ["g", "u", "b"],
    ),

    "w" : Action (
        name = "Whiff",
        
        cancelTimes = {
            "p" : 27,
            "k" : 27,
            "o" : 27,
            "t" : 10,
            "s" : 53,
            "a" : 25,
            "w" : 53,
            "g" : 1,
            "G" : 1,
            "u" : 1,
            "b" : 1
        },

        chargeActivations = { "s" : 53 },
        endActivations = ["g", "u", "b"],
        awaitCharges = { "s" : 0 }
    ),

    "g" : Action (
        name = "Get Over Here",
        damage = 25,
        
        damageTime = 17,
        maxTravelTime = 12,
        range = 20,
        cancelTimes = {
            "p" : 50,
            "k" : 50,
            "o" : 50,
            "t" : 46,
            "s" : 14,
            "a" : 14,
            "w" : 14,
            "g" : 102,
            "G" : 50,
            "u" : 47,
            "b" : 14
        },

        chargeActivations = { "g" : 102 },
        awaitCharges = { "g" : 0 }
    ),

    "G" : Action (
        name = "Get Over Here Targeting",
        damage = 55,

        damageTime = 37,
        maxTravelTime = 13,
        range = 24,
        cancelTimes = {
            "p" : 61,
            "k" : 61,
            "o" : 61,
            "t" : 61,
            "s" : 28,
            "a" : 28,
            "w" : 28,
            "g" : 61,
            "G" : 61,
            "u" : 32,
            "b" : 28
        },

        chargeActivations = { "g" : 61 },
        awaitCharges = { "g" : 0 }
    ),

    "u" : Action (
        name = "Uppercut",
        damage = 60, 
        procsTracer = True,
        
        damageTime = 23,
        range = 4,
        cancelTimes = {
            "p" : 48,
            "k" : 48,
            "o" : 48,
            "t" : 56,
            "s" : 19,
            "a" : 19,
            "w" : 19,
            "g" : 48,
            "G" : 48,
            "u" : 70,
            "b" : 19
        },

        chargeActivations = { "u" : 70 },
        awaitCharges = { "u" : 0 }
    ),

    "b" : Action (
        name = "Burn Tracer",
        damage = 30,
        
        damageTime = 17,
        range = 8,
        cancelTimes = {
            "p" : 30,
            "k" : 30,
            "o" : 30,
            "t" : 10,
            "s" : 10,
            "a" : 10,
            "w" : 10,
            "g" : 10,
            "G" : 10,
            "u" : 1,
            "b" : 70
        },

        chargeActivations = { "b" : 70 },
        awaitCharges = { "b" : 0 }
    )
}


# generate movestack timings procedurally to simplify code/maintenence
def loadMoveStacks(): 

    # ffamestack (G+u)
    for action in ACTIONS:
        ACTIONS[action].cancelTimes["G+u"] = ACTIONS[action].cancelTimes["G"]

    ffamestackCancelTimes = ACTIONS["G"].cancelTimes.copy()
    for key in ffamestackCancelTimes:
        ffamestackCancelTimes[key] = max(ACTIONS["u"].cancelTimes[key] + 1, ffamestackCancelTimes[key])

    ACTIONS["G+u"] = Action (
        name = "FFAmestack",
        damage = ACTIONS["G"].damage + ACTIONS["u"].damage,
        procsTracer = True,
        procTime = ACTIONS["u"].procTime + 1,
        damageTime = ACTIONS["G"].damageTime,
        firstDamageTime = ACTIONS["u"].firstDamageTime + 1,
        maxTravelTime = int(ACTIONS["G"].maxTravelTime * 4 / 20),
        cancelTimes = ffamestackCancelTimes,
        chargeActivations = {
            "g" : ACTIONS["G"].chargeActivations["g"],
            "u" : ACTIONS["u"].chargeActivations["u"] + 1
        },
        awaitCharges = { 
            "g" : 0,
            "u" : 1 
        }
    )

    # saporen (p+G, k+G, o+G)
    for action in ACTIONS:
        ACTIONS[action].cancelTimes["p+G"] = ACTIONS[action].cancelTimes["p"]
        ACTIONS[action].cancelTimes["k+G"] = ACTIONS[action].cancelTimes["k"]
        ACTIONS[action].cancelTimes["o+G"] = ACTIONS[action].cancelTimes["o"]

    punchSaporenCancelTimes = ACTIONS["G"].cancelTimes.copy()
    kickSaporenCancelTimes = ACTIONS["G"].cancelTimes.copy()
    overheadSaporenCancelTimes = ACTIONS["G"].cancelTimes.copy()
    for key in punchSaporenCancelTimes:
        punchSaporenCancelTimes[key] += ACTIONS["p"].cancelTimes["G"] - 1
        kickSaporenCancelTimes[key] += ACTIONS["k"].cancelTimes["G"] - 1
        overheadSaporenCancelTimes[key] += ACTIONS["o"].cancelTimes["G"]  -1

    ACTIONS["p+G"] = Action (
        name = "Punch Saporen",
        damage = ACTIONS["p"].damage + ACTIONS["G"].damage,
        procsTracer = True,
        procTime = ACTIONS["p"].procTime,
        damageTime = ACTIONS["p"].cancelTimes["G"] + ACTIONS["G"].damageTime - 1,
        firstDamageTime = ACTIONS["p"].damageTime,
        maxTravelTime = int(ACTIONS["G"].maxTravelTime * ACTIONS["p"].range / ACTIONS["G"].range),
        cancelTimes = punchSaporenCancelTimes,
        chargeActivations = {
            "g" : ACTIONS["p"].cancelTimes["G"] + ACTIONS["G"].chargeActivations["g"] - 1
        },
        awaitCharges = { 
            "g" :  ACTIONS["p"].cancelTimes["G"] - 1
        }
    )

    ACTIONS["k+G"] = Action (
        name = "Kick Saporen",
        damage = ACTIONS["k"].damage + ACTIONS["G"].damage,
        procsTracer = True,
        procTime = ACTIONS["k"].procTime,
        damageTime = ACTIONS["k"].cancelTimes["G"] + ACTIONS["G"].damageTime - 1,
        firstDamageTime = ACTIONS["k"].damageTime,
        maxTravelTime = int(ACTIONS["G"].maxTravelTime * ACTIONS["k"].range / ACTIONS["G"].range),
        cancelTimes = kickSaporenCancelTimes,
        chargeActivations = {
            "g" : ACTIONS["k"].cancelTimes["G"] + ACTIONS["G"].chargeActivations["g"] - 1
        },
        awaitCharges = { 
            "g" :  ACTIONS["k"].cancelTimes["G"] - 1
        }
    )

    ACTIONS["o+G"] = Action (
        name = "Overhead Saporen",
        damage = ACTIONS["o"].damage + ACTIONS["G"].damage,
        procsTracer = True,
        procTime = ACTIONS["o"].procTime,
        damageTime = ACTIONS["o"].cancelTimes["G"] + ACTIONS["G"].damageTime - 1,
        firstDamageTime = ACTIONS["o"].damageTime,
        maxTravelTime = int(ACTIONS["G"].maxTravelTime * ACTIONS["o"].range / ACTIONS["G"].range),
        cancelTimes = overheadSaporenCancelTimes,
        chargeActivations = {
            "g" : ACTIONS["o"].cancelTimes["G"] + ACTIONS["G"].chargeActivations["g"] - 1
        },
        awaitCharges = { 
            "g" :  ACTIONS["o"].cancelTimes["G"] - 1
        }
    )

    # saporen ffamestack (p+G+u, k+G+u, o+G+u)
    for action in ACTIONS:
        ACTIONS[action].cancelTimes["p+G+u"] = ACTIONS[action].cancelTimes["p"]
        ACTIONS[action].cancelTimes["k+G+u"] = ACTIONS[action].cancelTimes["k"]
        ACTIONS[action].cancelTimes["o+G+u"] = ACTIONS[action].cancelTimes["o"]

    punchSaporenFfamestackCancelTimes = ACTIONS["G+u"].cancelTimes.copy()
    kickSaporenFfamestackCancelTimes = ACTIONS["G+u"].cancelTimes.copy()
    overheadSaporenFfamestackCancelTimes = ACTIONS["G+u"].cancelTimes.copy()
    for key in punchSaporenFfamestackCancelTimes:
        punchSaporenFfamestackCancelTimes[key] += ACTIONS["p"].cancelTimes["G"] - 1
        kickSaporenFfamestackCancelTimes[key] += ACTIONS["k"].cancelTimes["G"] - 1
        overheadSaporenFfamestackCancelTimes[key] += ACTIONS["o"].cancelTimes["G"] - 1

    ACTIONS["p+G+u"] = Action (
        name = "Punch Saporen FFAmestack",
        damage = ACTIONS["p"].damage + ACTIONS["G"].damage + ACTIONS["u"].damage,
        procsTracer = True,
        procTime = ACTIONS["p"].procTime,
        damageTime = ACTIONS["p"].cancelTimes["G"] + ACTIONS["G+u"].damageTime -1,
        firstDamageTime = ACTIONS["p"].damageTime,
        maxTravelTime = int(ACTIONS["G"].maxTravelTime * ACTIONS["p"].range / ACTIONS["G"].range),
        cancelTimes = punchSaporenFfamestackCancelTimes,
        chargeActivations = {
            "g" : ACTIONS["p"].cancelTimes["G"] + ACTIONS["G"].chargeActivations["g"] - 1,
            "u" : ACTIONS["p"].cancelTimes["G"] + ACTIONS["u"].chargeActivations["u"]
        },
        awaitCharges = { 
            "g" :  ACTIONS["p"].cancelTimes["G"] - 1,
            "u" : ACTIONS["p"].cancelTimes["G"]
        }
    )

    ACTIONS["k+G+u"] = Action (
        name = "Kick Saporen FFAmestack",
        damage = ACTIONS["k"].damage + ACTIONS["G"].damage + ACTIONS["u"].damage,
        procsTracer = True,
        procTime = ACTIONS["k"].procTime,
        damageTime = ACTIONS["k"].cancelTimes["G"] + ACTIONS["G+u"].damageTime - 1,
        firstDamageTime = ACTIONS["k"].damageTime,
        maxTravelTime = int(ACTIONS["G"].maxTravelTime * ACTIONS["k"].range / ACTIONS["G"].range),
        cancelTimes = kickSaporenFfamestackCancelTimes,
        chargeActivations = {
            "g" : ACTIONS["k"].cancelTimes["G"] + ACTIONS["G"].chargeActivations["g"] - 1,
            "u" : ACTIONS["k"].cancelTimes["G"] + ACTIONS["u"].chargeActivations["u"]
        },
        awaitCharges = { 
            "g" :  ACTIONS["k"].cancelTimes["G"] - 1,
            "u" : ACTIONS["k"].cancelTimes["G"]
        }
    )

    ACTIONS["o+G+u"] = Action (
        name = "Overhead Saporen FFAmestack",
        damage = ACTIONS["o"].damage + ACTIONS["G"].damage + ACTIONS["u"].damage,
        procsTracer = True,
        procTime = ACTIONS["o"].procTime,
        damageTime = ACTIONS["o"].cancelTimes["G"] + ACTIONS["G+u"].damageTime - 1,
        firstDamageTime = ACTIONS["o"].damageTime,
        maxTravelTime = int(ACTIONS["G"].maxTravelTime * ACTIONS["o"].range / ACTIONS["G"].range),
        cancelTimes = overheadSaporenFfamestackCancelTimes,
        chargeActivations = {
            "g" : ACTIONS["o"].cancelTimes["G"] + ACTIONS["G"].chargeActivations["g"] - 1,
            "u" : ACTIONS["o"].cancelTimes["G"] + ACTIONS["u"].chargeActivations["u"]
        },
        awaitCharges = { 
            "g" :  ACTIONS["o"].cancelTimes["G"] - 1,
            "u" : ACTIONS["o"].cancelTimes["G"]
        }
    )

    # space jam (u+w+G)
    for action in ACTIONS:
        ACTIONS[action].cancelTimes["u+w+G"] = ACTIONS[action].cancelTimes["u"]

    spacejamCancelTimes = ACTIONS["G"].cancelTimes
    for key in spacejamCancelTimes:
        spacejamCancelTimes[key] += ACTIONS["u"].cancelTimes["s"] + ACTIONS["s"].cancelTimes["G"]

    ACTIONS["u+w+G"] = Action (
        name = "Space Jam",
        damage = ACTIONS["G"].damage + ACTIONS["u"].damage,
        procsTracer = True,
        procTime = ACTIONS["u"].procTime,
        damageTime =  ACTIONS["u"].cancelTimes["s"] + ACTIONS["s"].cancelTimes["G"] + ACTIONS["G"].damageTime,
        firstDamageTime = ACTIONS["u"].damageTime,
        maxTravelTime = int(ACTIONS["G"].maxTravelTime * ACTIONS["u"].range / ACTIONS["G"].range),
        cancelTimes = spacejamCancelTimes,
        chargeActivations = {
            "u" : ACTIONS["u"].cancelTimes["u"],
            "g" : ACTIONS["u"].cancelTimes["s"] + ACTIONS["s"].cancelTimes["G"] + ACTIONS["G"].chargeActivations["g"]
        },
        awaitCharges = { 
            "u" :  0,
            "s" : ACTIONS["u"].cancelTimes["s"],
            "g" : ACTIONS["u"].cancelTimes["s"]
        }
    )

    # reverse trigger (p+t, k+t, o+t) 
    for action in ACTIONS:
        ACTIONS[action].cancelTimes["p+t"] = ACTIONS[action].cancelTimes["p"]
        ACTIONS[action].cancelTimes["k+t"] = ACTIONS[action].cancelTimes["k"]
        ACTIONS[action].cancelTimes["o+t"] = ACTIONS[action].cancelTimes["o"]

    punchReverseTriggerCancelTimes = ACTIONS["t"].cancelTimes.copy()
    kickReverseTriggerCancelTimes = ACTIONS["t"].cancelTimes.copy()
    overheadReverseTriggerCancelTimes = ACTIONS["t"].cancelTimes.copy()
    for action in punchReverseTriggerCancelTimes:
        punchReverseTriggerCancelTimes[action] += ACTIONS["p"].damageTime - 1 - ACTIONS["t"].damageTime
        kickReverseTriggerCancelTimes[action] += ACTIONS["k"].damageTime - 1 - ACTIONS["t"].damageTime
        overheadReverseTriggerCancelTimes[action] += ACTIONS["o"].damageTime - 1 - ACTIONS["t"].damageTime

    ACTIONS["p+t"] = Action (
        name = "Punch Reverse Trigger",
        damage = ACTIONS["p"].damage + ACTIONS["t"].damage,
        procsTracer = True,
        procTime = 0,
        damageTime = ACTIONS["p"].damageTime,
        firstDamageTime = ACTIONS["p"].damageTime - 1,
        cancelTimes = punchReverseTriggerCancelTimes,
        chargeActivations = {
            "t" : ACTIONS["p"].damageTime - 1  - ACTIONS["t"].damageTime + ACTIONS["t"].chargeActivations["t"]
        },
        awaitCharges = { 
            "t" :  ACTIONS["p"].damageTime - 1 - ACTIONS["t"].damageTime
        }
    )

    ACTIONS["k+t"] = Action (
        name = "Kick Reverse Trigger",
        damage = ACTIONS["k"].damage + ACTIONS["t"].damage,
        procsTracer = True,
        procTime = 0,
        damageTime = ACTIONS["k"].damageTime,
        firstDamageTime = ACTIONS["k"].damageTime - 1,
        cancelTimes = kickReverseTriggerCancelTimes,
        chargeActivations = {
            "t" : ACTIONS["k"].damageTime - 1  - ACTIONS["t"].damageTime + ACTIONS["t"].chargeActivations["t"]
        },
        awaitCharges = { 
            "t" :  ACTIONS["k"].damageTime - 1 - ACTIONS["t"].damageTime
        }
    )

    ACTIONS["o+t"] = Action (
        name = "Overhead Reverse Trigger",
        damage = ACTIONS["o"].damage + ACTIONS["t"].damage,
        procsTracer = True,
        procTime = 0,
        damageTime = ACTIONS["o"].damageTime,
        firstDamageTime = ACTIONS["o"].damageTime - 1,
        cancelTimes = overheadReverseTriggerCancelTimes,
        chargeActivations = {
            "t" : ACTIONS["o"].damageTime - 1  - ACTIONS["t"].damageTime + ACTIONS["t"].chargeActivations["t"]
        },
        awaitCharges = { 
            "t" :  ACTIONS["o"].damageTime - 1 - ACTIONS["t"].damageTime
        }
    )

    # unique 3 hit (p+o, k+o)
    for action in ACTIONS:
        ACTIONS[action].cancelTimes["p+o"] = ACTIONS[action].cancelTimes["p"]
        ACTIONS[action].cancelTimes["k+o"] = ACTIONS[action].cancelTimes["k"]

    ACTIONS["p+o"] = Action (
        name = "Punch OH Stack",
        damage = ACTIONS["p"].damage + ACTIONS["o"].damage,
        procsTracer = True,
        procTime = ACTIONS["p"].procTime,
        damageTime = ACTIONS["o"].damageTime,
        firstDamageTime = ACTIONS["p"].damageTime,
        cancelTimes = ACTIONS["o"].cancelTimes.copy()
    )

    ACTIONS["k+o"] = Action (
        name = "Kick OH Stack",
        damage = ACTIONS["k"].damage + ACTIONS["o"].damage,
        procsTracer = True,
        procTime = ACTIONS["k"].procTime,
        damageTime = ACTIONS["o"].damageTime,
        firstDamageTime = ACTIONS["k"].damageTime,
        cancelTimes = ACTIONS["o"].cancelTimes.copy()
    )

    ACTIONS["j"] = Action (
        name = "Jump",
        cancelTimes = { a : 0 for a in ACTIONS }
    )

    ACTIONS["d"] = Action (
        name = "Double Jump",
        cancelTimes = { a : 0 for a in ACTIONS }
    )

    ACTIONS["l"] = Action (
        name = "Land",
        cancelTimes = { a : 0 for a in ACTIONS }
    )


from karen.evaluate import evaluate
from karen.getCombo import *

while True:

    inputString = input(">> ") + " " # space added so that commands with no arguments are easily handled
    while len(inputString) > 0 and inputString[0] in " !":
        inputString = inputString[1:]

    if len(inputString) > 4 and inputString[:5] == "help ":
        print("!eval [combo]: evaluates the given combo")
        print("!exit : closes the terminal")

    elif len(inputString) > 4 and inputString[:5] == "eval ":
        print(evaluate(inputString[5:], simpleMode=True))

    elif len(inputString) > 5 and inputString[:6] == "evala ":
        print(evaluate(inputString[6:]))

    elif len(inputString) > 5 and inputString[:6] == "evaln ":
        print(evaluate(inputString[6:], printWarnings=False))

    elif len(inputString) > 5 and inputString[:6] == "combo ":
        print(getCombo(inputString[6:]))

    elif len(inputString) > 5 and inputString[:7] == "combos ":
        print(listCombos())

    elif len(inputString) > 4 and inputString[:5] == "exit ":
        break

    else:
        print("Command not recognised, use '!help' to see a list of commands")


from setuptools import setup, find_packages
import pathlib

setup (
    name = "karen",
    version = "0.3.3a1",
    url = "https://github.com/EvilDuck14/Karen/",
    author = "EvilDuck",
    author_email = "theevilduck14@gmail.com",

    packages = find_packages(),
    install_requires = [],
    setup_requires=['setuptools-git'],

    description = "Evaluates & categorises Spider-Man's combos in Marvel Rivals.",
    long_description = (pathlib.Path(__file__).parent.resolve() / "README.md").read_text(encoding="utf-8"),
    long_description_content_type = "text/markdown",

    project_urls={
        "Discord" : "https://discord.gg/RpQf2zVAMP",
        "Source" : "https://github.com/EvilDuck14/Karen/",
    },
)


import discord
from discord.ext import commands

from karen.evaluate import evaluate
from karen.getCombo import *

import os
from dotenv import load_dotenv
load_dotenv()
BOT_TOKEN = os.getenv("BOT_TOKEN") # save your bot token as an environment variable or paste it here
print(BOT_TOKEN)

intents = discord.Intents.default()
intents.guild_messages = True
intents.message_content = True

bot = commands.Bot(command_prefix="!", intents=intents)

@bot.command()
async def eval(ctx, *arr):
    inputString = " ".join(str(x) for x in arr)
    output = evaluate(inputString, simpleMode=True)
    try:
        await ctx.send(output)
    except Exception as e:
        print(e)

@bot.command()
async def evala(ctx, *arr):
    inputString = " ".join(str(x) for x in arr)
    output = evaluate(inputString)
    try:
        await ctx.send(output)
    except Exception as e:
        print(e)

@bot.command()
async def evaln(ctx, *arr):
    inputString = " ".join(str(x) for x in arr)
    output = evaluate(inputString, printWarnings=False)
    try:
        await ctx.send(output)
    except Exception as e:
        print(e)

@bot.command()
async def combo(ctx, *arr):
    inputString = " ".join(str(x) for x in arr)
    output = getCombo(inputString)
    try:
        await ctx.send(output)
    except Exception as e:
        print(e)

@bot.command()
async def combos(ctx, *arr):
    output = listCombos()
    try:
        await ctx.send(output)
    except Exception as e:
        print(e)

@bot.event
async def on_ready():
    print(f'Logged in as {bot.user} (ID: {bot.user.id})')
    try:
        await bot.tree.sync()
        print("synced successfully")
    except Exception as e:
        print(e)
    

bot.run(BOT_TOKEN)


from karen.actions import *
from karen.state import State

def getComboSequence(inputString="", warnings=[]):

    if not ("G+u" in ACTIONS):
        loadMoveStacks()
    
    # removes anything in brackets from inputString string
    sequence = inputString
    while "(" in sequence and ")" in sequence[sequence.find("("):]:
        sequence = sequence[:sequence.find("(")] + sequence[sequence[sequence.find("("):].find(")") + sequence.find("(") + 1:]
    
    # removes leading/trailing spaces
    while sequence[0] == " ":
        sequence = sequence[1:]
    while sequence[-1] == " ":
        sequence = sequence[:-1]
    
    # uses long form if ">" is in sequence, there are invalid characters, or spaces
    if ">" in sequence or len([x for x in sequence if not (x.lower() in ACTION_NAMES or x in ["", " "])]) > 0 or len([x for x in sequence.split(" ")]) > 1:
        words = [x for x in sequence.replace("+", " + ").replace(">", " > ").split(" ") if x != ""]
        sequence = []
        while len(words) > 0:
            while words[0] == ">":
                words = words[1:]
            for j in range(0, len(words)):
                if ">" in words[0 : len(words) - j]: # ">" forces separation of actions
                    continue
                if ("".join(words[0 : len(words) - j])).lower() in ACTION_NAMES or len(words) - j == 1:
                    sequence += ["".join(words[0 : len(words) - j])]
                    words = words[len(words) - j:]
                    break
    
    # unrecognised actions
    warnings += [unknownAction + " is not a recognised action" for unknownAction in sequence if not unknownAction.lower() in ACTION_NAMES]
    
    # converts to a list of correctly formatted keys in ACTION_NAMES
    sequence = [(ACTION_NAMES[x] if x in ACTION_NAMES else x.lower()) for x in sequence if (x.lower() in ACTION_NAMES)] 
    sequence = list("".join(sequence))

    # folds movestack indicators into actions
    foldSequence = []
    for i in range(len(sequence)):
        if len(foldSequence) > 0 and sequence[i-1] == "+":
            foldSequence[-1] += "+" + sequence[i]
        elif sequence[i] != "+":
            foldSequence.append(sequence[i])
    
    # verify movestacks
    verifySequence = []
    for action in foldSequence:
        if action in ACTION_NAMES and ACTION_NAMES[action] in ACTIONS:
            verifySequence.append(ACTION_NAMES[action])
        else:
            warnings += [action + " is not a regocnised movestack"]
            verifySequence += action.split("+")

    return verifySequence

def simplify(sequence):
    newSequence = []

    for i in range(len(sequence) - 1):

        current = sequence[i]
        next = [j for j in sequence[i+1:] if not j in ["j", "d", "l"]][0]
        if next == "":
            newSequence += [current]
            continue

        # replace swing cancels with whiff cancels
        elif current == "s" and ACTIONS["w"].cancelTimes[next] <= ACTIONS["s"].cancelTimes[next]:
            newSequence += ["w"]

        # insert whiffs after actions with uppercuts where speed is increased
        elif "u" in current and ACTIONS[current].cancelTimes["w"] + ACTIONS["w"].cancelTimes[next] < ACTIONS[current].cancelTimes[next] and next != "":
            newSequence += [current, "w"]
        
        else:
            newSequence += [current]
    
    return newSequence + [""]
        

def addAction(state=State(), action="", nextAction="", warnings=[], maxTravelTimes=False, simpleMode=False):

    if not ("G+u" in ACTIONS):
        loadMoveStacks()

    # awaits required cooldowns
    for a in ACTIONS[action].awaitCharges:
        state.incrementTime(state.charges[a].activeTimer, warnings) 
        state.incrementTime(state.charges[a].cooldownTimer - ACTIONS[action].awaitCharges[a], warnings)
        state.incrementTime(state.charges[a].RECHARGE_TIME - state.charges[a].currentCharge - ACTIONS[action].awaitCharges[a], warnings)

    # awaits tracer register for GOHT
    if "g" in ACTIONS[action].awaitCharges:
        state.incrementTime(state.gohtWaitTime - ACTIONS[action].awaitCharges["g"], warnings)

    # awaits kick expiration for punch
    if "p" in action and state.punchSequence == 2:
        state.incrementTime(state.punchSequenceTimer, warnings)

    if "k" in action and state.punchSequence < 2:
        warnings += ["uses impossible kick after " + state.sequence]
    
    if "G" in action and (state.tracerActiveTimer == 0 or state.tracerActiveTimer < ACTIONS[action].awaitCharges["g"]) and (state.burnTracerActiveTimer == 0 or state.burnTracerActiveTimer < ACTIONS[action].awaitCharges["g"]):
        warnings += ["uses GOHT on nonxistent or expired tracer after " + state.sequence]

    if "p" in action and (state.hasSwingOverhead or state.hasJumpOverhead) and not simpleMode:
        warnings += ["uses punch when overhead was expected after " + state.sequence]
    if "k" in action and (state.hasSwingOverhead or state.hasJumpOverhead and not simpleMode):
        warnings += ["uses kick when overhead was expected after " + state.sequence]

    # awaits whiff end for overhead
    if "o" in action and (not state.hasJumpOverhead) and (not state.hasSwingOverhead) and state.charges["s"].activeTimer > 0:
        print(f"TIME ELAPSED: {state.timeTaken}")
        print(f"ACTIVE TIMER: {state.charges["s"].activeTimer}")
        state.incrementTime(state.charges["s"].activeTimer, warnings)

    # awaits punch timer for punch
    if "p" in action or "k" in action:
        state.incrementTime(state.punchWaitTimer, warnings)

    # awaits swing timer for swing/whiff
    if "s" in action or "w" in action:
        state.incrementTime(state.swingWaitTimer - ACTIONS[action].awaitCharges["s"], warnings)

    # precomputes max travel time, factoring in the known max range
    travelTime = 0 if not maxTravelTimes else ACTIONS[action].maxTravelTime
    if action != "s" and ACTIONS[action].range > state.maxPossibleRange:
        travelTime = int(travelTime / ACTIONS[action].range * state.maxPossibleRange)
    if action == "s" and nextAction == "":
        travelTime = 0

    # handling changes of max possible range
    if action != "s" and ACTIONS[action].range != 0:
        state.maxPossibleRange = min(state.maxPossibleRange, ACTIONS[action].range)
    if action == "b":
        state.maxPossibleRange += BURN_TRACER_BACKFLIP_DISTANCE
    if action == "G":
        state.maxPossibleRange = 0

    # punch sequence increment
    if "p" in action:
        state.punchSequence += 1
        state.punchSequenceTimer = PUNCH_SEQUENCE_MAX_DELAY
    if "k" in action:
        state.punchSequence = 0
    
    # processes overhead logic
    if "o" in action and (not state.hasSwingOverhead) and (not state.hasJumpOverhead) and not simpleMode:
        warnings += ["uses impossible overhead after " + state.sequence]

    if action == "l":
        state.isAirborn = False
        state.hasDoubleJump = True
        state.hasSwingOverhead = False
        state.hasJumpOverhead = False
   
    if action == "j" and state.isAirborn:
        if not state.hasDoubleJump and not simpleMode: 
            warnings += ["uses impossible double jump after " + state.sequence]
        state.hasDoubleJump = False
        state.hasJumpOverhead = True

    elif action in ["j", "s", "a", "b"] or "u" in action:
        state.isAirborn = True

    if action == "d":
        if not state.hasDoubleJump and not simpleMode: 
            warnings += ["uses impossible double jump after " + state.sequence]
        state.isAirborn = True
        state.hasDoubleJump = False
        state.hasJumpOverhead = True

    if action in ["o", "G", "G+u", "p+G", "k+G", "p+G+u", "k+G+u", "p+G+u", "k+G+u", "o+t", "p+o", "k+o"]:
        if state.hasSwingOverhead:
            state.hasSwingOverhead = False
        else:
            state.hasJumpOverhead = False
    if action in ["o+G", "o+G+u"]:
        state.hasSwingOverhead = False
        state.hasJumpOverhead = False
    if action == "u+w+G":
        state.hasSwingOverhead = True
        state.hasJumpOverhead = False

    if action in ["s", "a", "b"] or "u" in action:
        state.hasDoubleJump = True
        state.hasJumpOverhead = False
    
    if action == "b":
        state.hasSwingOverhead = True

    # ends current cancellable actions
    for cancelCharge in ACTIONS[action].endActivations:
        if state.charges[cancelCharge].activeTimer > 0:
            state.endAction(cancelCharge)

    # activating actions/consuming cooldowns
    if action in ["s", "a"]:
        state.removeSwingOnEnd = True
    if "w" in action:
        state.removeSwingOnEnd = False

    for charge in ACTIONS[action].chargeActivations:
        state.charges[charge].cooldownTimer = state.charges[charge].COOLDOWN_TIME
        if ACTIONS[action].chargeActivations[charge] == 0:
            state.charges[charge].currentCharge -= state.charges[charge].RECHARGE_TIME
        else:
            state.charges[charge].activeTimer = ACTIONS[action].chargeActivations[charge] + travelTime
        
    # adding tracer tags
    if action == "t":
        state.tracerActiveTimer = TRACER_ACTIVE_TIME + ACTIONS["t"].damageTime + travelTime
        if state.tracerActiveTimer == 0 and state.burnTracerActiveTimer == 0:
            state.gohtWaitTime = ACTIONS[action].damageTime + (0 if not maxTravelTimes else TRACER_MAX_TRAVEL_TIME)
    if action == "b":
        state.burnTracerActiveTimer = BURN_TRACER_ACTIVE_TIME + ACTIONS["b"].damageTime + travelTime  
        if state.tracerActiveTimer == 0 and state.burnTracerActiveTimer == 0:
            state.gohtWaitTime = ACTIONS[action].damageTime + (0 if not maxTravelTimes else BURN_TRACER_MAX_TRAVEL_TIME)

    # proccing tracers
    if ACTIONS[action].procsTracer and state.tracerActiveTimer >= ACTIONS[action].procTime or action in ["p+t", "k+t", "o+t"]:
        state.damageDealt += TRACER_PROC_DAMAGE
        state.tracerActiveTimer = 0
    if ACTIONS[action].procsTracer and state.burnTracerActiveTimer >= ACTIONS[action].procTime:
        state.burnActiveTimer = BURN_TRACER_BURN_TIME + ACTIONS[action].procTime
        state.burnTracerActiveTimer = 0

    if state.firstDamageTime == 0 and ACTIONS[action].damage > 0:
        state.firstDamageTime = state.timeTaken + ACTIONS[action].firstDamageTime + (travelTime if not "+G" in action else 0)

    # autoswing delays next swing/whiff
    if action == "a":
        state.swingWaitTimer = ACTIONS["a"].cancelTimes["s"]

    state.damageDealt += ACTIONS[action].damage
    state.minTimeTaken = max(state.minTimeTaken, state.timeTaken + ACTIONS[action].damageTime + travelTime)
    state.incrementTime((ACTIONS[action].damageTime if nextAction == "" else ACTIONS[action].cancelTimes[nextAction]) + travelTime, warnings)

    state.sequence += ("" if state.sequence == "" else " > ") + ACTIONS[action].name


from karen.state import State
from karen.combo import *
from karen.classify import classify
from math import floor
from karen.actions import *

def evaluate(inputString, printWarnings = True, simpleMode = False):
    warnings = []

    state = State()
    comboSequence = getComboSequence(inputString, warnings) + [""]

    # runs a second evaluation with maximum action ranges
    maxTravelTimeState = State()
    maxTravelTimeState.inferInitialState(comboSequence)

    # infer initial state being airborne/having overhead
    state.inferInitialState(comboSequence, warnings)
    comboSequence = state.correctSequence(comboSequence)

    if simpleMode:
        comboSequence = simplify(comboSequence)

    for i in range(len(comboSequence) - 1):
        nextAction = [j for j in comboSequence[i+1:] if not j in ["j", "d", "l"]][0]
        addAction(state, comboSequence[i], nextAction, warnings, simpleMode=simpleMode)
        addAction(maxTravelTimeState, comboSequence[i], nextAction, maxTravelTimes=True)
    state.resolve(warnings)
    maxTravelTimeState.resolve()

    # checks for continued burn tracer damage after final action
    burnTracerBonusDamage = ""
    if state.burnActiveTimer > 12:
        burnTracerBonusDamage = "(plus " + str(int(floor((state.burnActiveTimer - 1) / 12) * BURN_TRACER_DPS / 5)) +" burn over time)"

    comboName = classify("".join(comboSequence))

    timeSecondsMin = round(state.timeTaken / 60, 2)
    timeSecondsMax = round(maxTravelTimeState.timeTaken / 60, 2)
    showTimeRange = maxTravelTimeState.timeTaken != state.timeTaken and not simpleMode
    timeSeconds = str(timeSecondsMin) + (f"-{timeSecondsMax}" if showTimeRange else "")
    timeFrames = str(state.timeTaken) + (f"-{maxTravelTimeState.timeTaken}" if showTimeRange else "")

    timeFromDamageSecondsMin = round(state.timeFromDamage / 60, 2)
    timeFromDamageSecondsMax = round(maxTravelTimeState.timeFromDamage / 60, 2)
    showTimeFromDamageRange = maxTravelTimeState.timeFromDamage != state.timeFromDamage and not simpleMode
    timeFromDamageSeconds = str(timeFromDamageSecondsMin) + (f"-{timeFromDamageSecondsMax}" if showTimeFromDamageRange else "")
    timeFromDamageFrames = str(state.timeFromDamage) + (f"-{maxTravelTimeState.timeFromDamage}" if showTimeFromDamageRange else "")

    output = ( f"### {comboName}"
    f"\n> {state.sequence}"
    f"\n**Time:** {timeSeconds} seconds ({timeFrames} frames)"
    f"{f"\n**Time From Damage:** {timeFromDamageSeconds} seconds ({timeFromDamageFrames} frames)" if state.damageDealt > 0 else ""}"
    f"\n**Damage:** {int(state.damageDealt)} {burnTracerBonusDamage}" )

    if len(warnings) > 0 and printWarnings:
        warninglist = ["\nWARNING: " + x for x in warnings]

        output += "\n```"
        while len(warninglist) > 0 and len(output) + len(warninglist[0]) + len(f"\n...({len(warninglist)} warnings)\n```") < 2000:
            output += warninglist[0]
            warninglist = warninglist[1:]
        if len(warninglist) > 0:
            output += f"\n...({len(warninglist)} warnings)"
        output += "\n```"

    if len(output) >= 2000:
        return "```\nERROR: Combo too long for Discord API\n```"  
    
    return output


CLASSIFICATIONS = {
    "tGu" : "Bread & Butter (BnB)",
    "tGuwtt" : "BnB Tracers",
    "tGuso" : "BnB Short Plink",
    "tGuwto" : "BnB Long Plink",
    "tGupt" : "BnB Punch",
    "tGutp" : "Reverse Panther",
    "tGuwpp+o" : "BnB U3H",
    "tGwtuwt" : "Weave Combo",
    "tGptu" : "Panther Combo",
    "tGsptu" : "Fast Panther",
    "tptu" : "One Two",
    "totu" : "Overhead One Two",
    "tgptu" : "Fishing Combo / Sekkombo",
    "tgotu" : "Reverse Yo-Yo",
    "tgktu" : "Grip Kick Rip (GKR)",
    "otG+u" : "Fast FFAmestack",
    "otuwtg" : "Overhead Burst",
    "otsptu" : "Master Manipulator",
    "otgwtu" : "Master Masher",
    "otgwtuwto" : "Fantastic Killer",
    "otuwtotgwtuwtoto" : "Further Beyond",
    "to+G+u" : "Saporen FFAmestack",
    "gwtuwto" : "Yo-Yo",
    "gwtuototu" : "Botched Yo-Yo",
    "gtutp" : "Agni-Kai Yo-Yo",
    "tuwtGu" : "Driveby",
    "twGuot" : "Bald Slam",
    "uwtGo" : "Vortex",
    "tgptsptu" : "Hydro Combo",

    "tGbu" : "Burn BnB / Fadeaway",
    "otbu" : "Burn Overhead Burst",
    "tgptbu" : "Fried Fish / Firehook",
    "tbuwtg" : "In And Out",
    "tptbu" : "Burn One Two",
    "tGbtu" : "Burn Weave",
    "tGwtbu" : "Burn Weave"
}

def classify(comboString):
    comboString = comboString.replace("j", "").replace("d", "").replace("l", "") # doesnt consider jumping/landing in classificaiton
    comboString = comboString.replace("a", "s") # autoswings are equivalent to swings for classification purposes

    if comboString in CLASSIFICATIONS:
        return CLASSIFICATIONS[comboString]
    
    # adding tracer to the end is considered the same combo
    if len(comboString) >= 1 and comboString[-1] == "t" and comboString[:-1] in CLASSIFICATIONS:
        return CLASSIFICATIONS[comboString[:-1]]
    if len(comboString) >= 2 and comboString[-2:] == "wt" and comboString[:-2] in CLASSIFICATIONS:
        return CLASSIFICATIONS[comboString[:-2]]
    
    # attempts classification assuming swing cancels are typos of whiff cancels
    if "st" in comboString:
        attempt = classify(comboString.replace("st", "wt")) 
        if attempt != "Undocumented":
            return "Wasteful " + attempt
    
    # attempts to classify as a slow variation of another combo by adding whiff cancels
    improveString = comboString.replace("ut", "uwt").replace("gt", "gwt").replace("ug", "uwg")
    if improveString != comboString:
        attempt = classify(improveString)
        if attempt != "Undocumented":
            return "Slow " + attempt
    
    return "Undocumented"